/*
 *  Copyright (C) 2004, 2005, 2006, 2007  James Antill
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  email: james@and.org
 */
/* conditionally compliant HTTP/1.1 server. */
#define _GNU_SOURCE 1 /* strsignal() / posix_fadvise64 */
#include <vstr.h>

#include <socket_poll.h>
#include <timer_q.h>

#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <err.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#include <grp.h>

#include <sys/wait.h>

#define EX_UTILS_NO_USE_INIT  1
#define EX_UTILS_NO_USE_EXIT  1
#define EX_UTILS_NO_USE_LIMIT 1
#define EX_UTILS_NO_USE_OPEN  1
#define EX_UTILS_NO_USE_GET   1
#define EX_UTILS_NO_USE_IO_FD 1
#define EX_UTILS_RET_FAIL     1
#include "ex_utils.h"

#include "mk.h"

MALLOC_CHECK_DECL();

#include "cntl.h"
#include "date.h"

#define AUDISPD_HAVE_GLOBAL_OPTS 1
#include "auditd_audispd_conf.h"

#include <syslog.h>

AUDISPD_CONF_DECL_OPTS(audispd_opts);

#ifndef TCP_CONGESTION
# ifdef __linux__
#  define TCP_CONGESTION 13
# else
#  define TCP_CONGESTION 0
# endif
#endif

/* we get prctl.h from evnt.h */
#if defined(HAVE_SYS_CAPABILITY_H) && defined(PR_SET_KEEPCAPS)
# include <sys/capability.h>
# define PROC_CNTL_KEEPCAPS(x1) prctl(PR_SET_KEEPCAPS, x1, 0, 0, 0)
#else
# define PROC_CNTL_KEEPCAPS(x1) (errno = ENOSYS, -1)
# define cap_t void *
# define cap_from_text(x) (errno = ENOSYS, NULL)
# define cap_set_proc(x) (errno = ENOSYS, -1)
# define cap_free(x) (errno = ENOSYS, -1)
#endif

#ifdef PR_SET_DUMPABLE
# define PROC_CNTL_DUMPABLE(x1) prctl(PR_SET_DUMPABLE, x1, 0, 0, 0)
#else
# define PROC_CNTL_DUMPABLE(x1) (errno = ENOSYS, -1)
#endif




#define CLEN COMPILE_STRLEN

/* is the cstr a prefix of the vstr */
#define VPREFIX(vstr, p, l, cstr)                                       \
    (((l) >= CLEN(cstr)) &&                                             \
     vstr_cmp_buf_eq(vstr, p, CLEN(cstr), cstr, CLEN(cstr)))
/* is the cstr a suffix of the vstr */
#define VSUFFIX(vstr, p, l, cstr)                                       \
    (((l) >= CLEN(cstr)) &&                                             \
     vstr_cmp_eod_buf_eq(vstr, p, l, cstr, CLEN(cstr)))

/* is the cstr a prefix of the vstr, no case */
#define VIPREFIX(vstr, p, l, cstr)                                      \
    (((l) >= CLEN(cstr)) &&                                             \
     vstr_cmp_case_buf_eq(vstr, p, CLEN(cstr), cstr, CLEN(cstr)))

/* for simplicity */
#define VALLEQ(vstr, cstr) vstr_cmp_cstr_eq(vstr, 1, (vstr)->len, cstr)


static Vlg *vlg = NULL;

static void usage(const char *program_name, int ret, const char *prefix)
{
  Vstr_base *out = vstr_make_base(NULL);

  if (!out)
    errno = ENOMEM, err(EXIT_FAILURE, "usage");

  vstr_add_fmt(out, 0, "%s\n\
 Format: %s [options]\n\
  Options\n\
    --configuration-file\n\
    --config-file  -C - Load configuration file specified.\n\
    --plugins         - Load <dir>/*.conf as plugin configuration files.\n\
    --policies        - Load <dir>/*.conf as policy configuration files.\n\
    --debug -d        - Raise debug level (can be used upto 3 times).\n\
    --help -h         - Print this message.\n\
    --version -V      - Print the version string.\n\
",
               prefix, program_name);

  if (io_put_all(out, ret ? STDERR_FILENO : STDOUT_FILENO) == IO_FAIL)
    err(EXIT_FAILURE, "write");
  
  exit (ret);
}

/* HACK ... */
extern const char *audit_msg_type_to_name(int msg_type);

static int audisp__fmt_add_msg_type2name(Vstr_base *base, size_t pos,
                                         Vstr_fmt_spec *spec)
{
  unsigned int type = VSTR_FMT_CB_ARG_VAL(spec, unsigned int, 0);

  return (vstr_add_sysfmt(base, pos, "%s", audit_msg_type_to_name(type)));
}

static int audisp_fmt_add_msg_type2name(Vstr_conf *conf, const char *name)
{
  return (vstr_fmt_add(conf, name, audisp__fmt_add_msg_type2name,
                       VSTR_TYPE_FMT_UINT,
                       VSTR_TYPE_FMT_END));

}

static void serv_init(void)
{
  if (!vstr_init()) /* init the library */
    errno = ENOMEM, err(EXIT_FAILURE, "init");

  vlg_init();

  if (!vstr_cntl_conf(NULL, VSTR_CNTL_CONF_SET_TYPE_GRPALLOC_CACHE,
                      VSTR_TYPE_CNTL_CONF_GRPALLOC_CSTR))
    errno = ENOMEM, err(EXIT_FAILURE, "init");

  if (!vstr_cntl_conf(NULL,
                      VSTR_CNTL_CONF_SET_NUM_BUF_SZ, OPT_SERV_CONF_BUF_SZ))
    errno = ENOMEM, err(EXIT_FAILURE, "init");
  
  /* no passing of conf to evnt */
  if (!vstr_cntl_conf(NULL, VSTR_CNTL_CONF_SET_FMT_CHAR_ESC, '$') ||
      !vstr_cntl_conf(NULL, VSTR_CNTL_CONF_SET_LOC_CSTR_THOU_SEP, "_") ||
      !vstr_cntl_conf(NULL, VSTR_CNTL_CONF_SET_LOC_CSTR_THOU_GRP, "\3") ||
      !vstr_sc_fmt_add_all(NULL) ||
      !vlg_sc_fmt_add_all(NULL) ||
      !VSTR_SC_FMT_ADD(NULL, audisp_fmt_add_msg_type2name,
                       "<au-msg_type2name", "u", ">") /* ||
       !VSTR_SC_FMT_ADD(NULL, http_fmt_add_vstr_add_vstr,
                       "<http-esc.vstr", "p%zu%zu", ">") ||
      !VSTR_SC_FMT_ADD(NULL, http_fmt_add_vstr_add_sect_vstr,
      "<http-esc.vstr.sect", "p%p%u", ">") */ )
    errno = ENOMEM, err(EXIT_FAILURE, "init");

  if (!(vlg = vlg_make()))
    errno = ENOMEM, err(EXIT_FAILURE, "init");

  if (!VSTR_SC_FMT_ADD(vlg->out_vstr->conf, audisp_fmt_add_msg_type2name,
                       "<au-msg_type2name", "u", ">"))
    errno = ENOMEM, err(EXIT_FAILURE, "init");
  
  /*
  if (!VSTR_SC_FMT_ADD(vlg->out_vstr->conf, http_fmt_add_vstr_add_vstr,
                       "<http-esc.vstr", "p%zu%zu", ">") ||
      !VSTR_SC_FMT_ADD(vlg->out_vstr->conf, http_fmt_add_vstr_add_sect_vstr,
                       "<http-esc.vstr.sect", "p%p%u", ">"))
    errno = ENOMEM, err(EXIT_FAILURE, "init");
  if (!VSTR_SC_FMT_ADD(vlg->sig_out_vstr->conf, http_fmt_add_vstr_add_vstr,
                       "<http-esc.vstr", "p%zu%zu", ">") ||
      !VSTR_SC_FMT_ADD(vlg->sig_out_vstr->conf, http_fmt_add_vstr_add_sect_vstr,
                       "<http-esc.vstr.sect", "p%p%u", ">"))
    errno = ENOMEM, err(EXIT_FAILURE, "init");
  */
  
  if (!socket_poll_init(0, SOCKET_POLL_TYPE_MAP_DIRECT))
      errno = ENOMEM, err(EXIT_FAILURE, "init");
  
  evnt_logger(vlg);
  evnt_timeout_init();

  vlg_time_set(vlg, evnt_sc_time);
  
  opt_serv_logger(vlg);

  /* httpd_init(vlg); */
  
  opt_serv_sc_signals();
}

static struct Evnt *audispd__input = NULL;
static unsigned int audispd__io_num      = 0;
static unsigned int audispd__io_done_num = 0;
static struct Con *con__beg = NULL;

static int audispd__serv_send_err(struct Con *con, const char *msg)
{
  if (errno != EPIPE)
    vlg_warn(vlg, "send(%s): %m\n", msg);
  else
    vlg_dbg2(vlg, "send(%s): SIGPIPE $<sa:%p>\n", msg, CON_CEVNT_SA(con));

  return (FALSE);
}

static int audispd_serv_send(struct Con *con)
{ /* FIXME: opt? ... shared with and-httpd */
  if (!con->evnt->io_w->len)
    return (TRUE);
  
  ASSERT(!con->evnt->io_w->conf->malloc_bad);
  while (con->evnt->io_w->len)
  {
    if (!con->io_limit_num--)
    {
      vlg_dbg2(vlg, "audispd Q send $<sa:%p>\n", CON_CEVNT_SA(con));
      if (!evnt_send_add(con->evnt, TRUE, AUDISPD_CONF_MAX_WAIT_SEND))
        return (audispd__serv_send_err(con, "Q"));
      
      /* queued */
      break;
    }
    
    if (!evnt_send(con->evnt))
      return (audispd__serv_send_err(con, "end"));
  }

  if (con->flag_drop_msgs)
    con->flag_drop_msgs = (con->evnt->io_w->len > con->policy->max_data);
  
  if (!con->evnt->io_w->len)
  {
    if (!audispd__input)
      return (FALSE);
    
    ++audispd__io_done_num;
    if (audispd__io_done_num == audispd__io_num)
      vstr_del(audispd__input->io_w, 1, audispd__input->io_w->len);
  }
  
  return (TRUE);
}

static int audispd_serv_recv(struct Con *con)
{
  unsigned int ern = 0;
  int ret = FALSE;

  /* FIXME: opt? ... mostly shared with and-httpd */
  ASSERT(!con->evnt->io_r_shutdown);
  
  if (!con->io_limit_num--)
    return (TRUE);

  if (!(ret = evnt_recv(con->evnt, &ern)))
  {
    if (ern != VSTR_TYPE_SC_READ_FD_ERR_EOF)
    {
      vlg_dbg2(vlg, "RECV ERR from[$<sa:%p>]: %u\n", CON_CEVNT_SA(con), ern);
      goto con_cleanup;
    }
    if (!evnt_shutdown_r(con->evnt, TRUE))
      goto con_cleanup;
  }

  /* if (audispd_parse_plugin_cmds())
   *   return (TRUE);
   */

  /* FIXME: just ignore everything atm. */
  vstr_del(con->evnt->io_r, 1, con->evnt->io_r->len);
  
  return (TRUE);
  
 con_cleanup:
  con->evnt->io_r->conf->malloc_bad = FALSE;
  con->evnt->io_w->conf->malloc_bad = FALSE;
    
  return (FALSE);
}

static int serv_cb_func_send(struct Evnt *evnt)
{
  struct Con *con = (struct Con *)evnt;
  
  assert(AUDISPD_CONF_SEND_CALL_LIMIT >= 1);
  con->io_limit_num = AUDISPD_CONF_SEND_CALL_LIMIT;
  return (audispd_serv_send(con));
}

static int serv_cb_func_recv(struct Evnt *evnt)
{
  struct Con *con = (struct Con *)evnt;
  
  assert(AUDISPD_CONF_RECV_CALL_LIMIT >= 1);
  con->io_limit_num = AUDISPD_CONF_RECV_CALL_LIMIT;
  return (audispd_serv_recv(con));
}

static int audispd_con_init_beg(struct Con *con)
{
  if (!con || !(con->tag = vstr_dup_cstr_ptr(NULL, "<default>")))
    return (FALSE);
  
  con->flag_drop_msgs = FALSE;
  con->managed = NULL;

  ++audispd__io_num;
  con->prev = NULL;
  if ((con->next = con__beg))
    con->next->prev = con;
  con__beg = con;

  return (TRUE);
}

static int audispd_con_init_end(struct Con *con, 
                                struct Acpt_listener *acpt_listener)
{
  Audispd_policy_opts *po = (Audispd_policy_opts *)audispd_opts->s->def_policy;
  int ret = TRUE;

  if (acpt_listener)
  {
    con->evnt->acpt_sa_ref = vstr_ref_add(acpt_listener->ref);
  
    if (acpt_listener->def_policy)
      po = acpt_listener->def_policy->ptr;
  }
  
  audispd_policy_change_con(con, po);
  
  /* if (!audispd_policy_connection(con, audispd_opts->conf,
                                 audispd_opts->match_connection))
  {
    Vstr_base *s1 = con->policy->s->policy_name;
    Vstr_base *s2 = audispd_opts->conf->tmp;
    
    vlg_info(vlg, "CONF-MAIN-ERR from[$<sa:%p>]: policy $<vstr.all:%p>"
             " backtrace: $<vstr.all:%p>\n", CON_CEVNT_SA(con), s1, s2);
    ret = FALSE;
  }
  else */ if (con->evnt->flag_q_closed)
  {
    Vstr_base *s1 = con->policy->s->policy_name;
    vlg_info(vlg, "BLOCKED from[$<sa:%p>]: policy $<vstr.all:%p>\n",
             CON_CEVNT_SA(con), s1);
    ret = FALSE;
  }
  
  return (ret);
}

/* Fwd ref: _free calls this, which calls make if managed... */
static const struct Evnt_cbs audispd__con_evnt_cbs;

static int serv_make_managed_con(Audispd_cmd *cmd)
{
  Opt_serv_opts *opts = audispd_opts->s;
  const char *name = NULL;
  const char *path = NULL;
  const char *arg1 = NULL;
  const char *arg2 = NULL;
  const char *arg3 = NULL;
  const char *arg4 = NULL;
  const char *arg5 = NULL;
  const char *arg6 = NULL;
  const char *arg7 = NULL;
  const char *arg8 = NULL;
  int fds[2];
  pid_t pid = -1;
  struct Con *con = NULL;

  /* Note: this must work, as it's already been done. */
  name = vstr_export_cstr_ptr(cmd->name, 1, cmd->name->len);
  path = vstr_export_cstr_ptr(cmd->path, 1, cmd->path->len);

  if (cmd->arg_num > 0)
    arg1 = vstr_export_cstr_ptr(cmd->args[0], 1, cmd->args[0]->len);
  if (cmd->arg_num > 1)
    arg2 = vstr_export_cstr_ptr(cmd->args[1], 1, cmd->args[1]->len);
  if (cmd->arg_num > 2)
    arg3 = vstr_export_cstr_ptr(cmd->args[2], 1, cmd->args[2]->len);
  if (cmd->arg_num > 3)
    arg4 = vstr_export_cstr_ptr(cmd->args[3], 1, cmd->args[3]->len);
  if (cmd->arg_num > 4)
    arg5 = vstr_export_cstr_ptr(cmd->args[4], 1, cmd->args[4]->len);
  if (cmd->arg_num > 5)
    arg6 = vstr_export_cstr_ptr(cmd->args[5], 1, cmd->args[5]->len);
  if (cmd->arg_num > 6)
    arg7 = vstr_export_cstr_ptr(cmd->args[6], 1, cmd->args[6]->len);
  if (cmd->arg_num > 7)
    arg8 = vstr_export_cstr_ptr(cmd->args[7], 1, cmd->args[7]->len);
  
  if (socketpair(PF_LOCAL, SOCK_STREAM, 0, fds) == -1)
    VLG_WARN_GOTO(sockpair_fail, (vlg, "socketpair(): %m\n"));
  
  switch ((pid = fork()))
  {
    case -1: VLG_WARN_GOTO(fork_fail, (vlg, "fork(): %m\n"));
      
    case 0: /* child */
      close(fds[0]);
      if (dup2(fds[1], STDIN_FILENO) == -1)
        err(EXIT_FAILURE, "dup2(STDIN)");
      if (dup2(fds[1], STDOUT_FILENO) == -1)
        err(EXIT_FAILURE, "dup2(STDOUT)");
      close(fds[1]);
      execl(path, path, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8,
            (char *)NULL);
      abort();
      break;
      
    default:
      break;
  }
  
  /* FIXME: dup code */
  if (!(con = MK(sizeof(struct Con))))
    VLG_WARNNOMEM_GOTO(mk_con_fail, (vlg, "make con: %m\n"));
  
  if (!audispd_con_init_beg(con))
    VLG_WARNNOMEM_GOTO(init_con_fail, (vlg, "init con: %m\n"));
  con->managed = cmd;
  
  { /* FIXME: something better, for SA names? */
    struct sockaddr_un tmp_sun;
    struct sockaddr_un *saddr = NULL;
    Vstr_ref *sa = NULL;
    int ret = FALSE;
    
    tmp_sun.sun_path[0] = 0;
    
    if (!(sa = vstr_ref_make_malloc(SUN_LEN(&tmp_sun) + 1 +
                                    strlen("managed-connection-") +
                                    strlen(name))))
      VLG_WARNNOMEM_GOTO(mk_sa_fail, (vlg, "make sa: %m\n"));
    saddr = sa->ptr;
    saddr->sun_family = AF_LOCAL;
    strcpy(saddr->sun_path, "managed-connection-");
    strcat(saddr->sun_path, name);
    
    ret = evnt_make_custom(con->evnt, fds[0], sa, POLLIN);
    vstr_ref_del(sa);

    if (!ret)
      VLG_WARNNOMEM_GOTO(mk_evnt_fail, (vlg, "make evnt: %m\n"));

    con->evnt->cbs = &audispd__con_evnt_cbs;

    close(fds[1]);
  }
  
  if (!evnt_limit_dup(con->evnt, &opts->io_nslimit))
    VLG_WARNNOMEM_GOTO(mk_limit_fail, (vlg, "make limit: %m\n"));
  
  if (!evnt_limit_add(con->evnt,  opts->ref_io_limit))
    VLG_WARNNOMEM_GOTO(mk_limit_fail, (vlg, "make limit: %m\n"));
  
  if (!evnt_limit_dup(con->evnt, &opts->io_nslimit))
    VLG_WARNNOMEM_GOTO(mk_limit_fail, (vlg, "make limit: %m\n"));
  
  if (!evnt_limit_add(con->evnt,  opts->ref_io_limit))
    VLG_WARNNOMEM_GOTO(mk_limit_fail, (vlg, "make limit: %m\n"));
  
  if (!audispd_con_init_end(con, NULL))
    VLG_WARNNOMEM_GOTO(init_con_end_fail, (vlg, "init con end: %m\n"));
  
  if (!evnt_sc_timeout_via_mtime(con->evnt,
                                 con->policy->s->idle_timeout * 1000))
    VLG_WARNNOMEM_GOTO(mk_tm_fail, (vlg, "make tm: %m\n"));
  
  con->evnt->flag_io_wonly_tm = TRUE;
  
  if (cmd->def_policy->len)
  {
    Vstr_base *defp = cmd->def_policy;
    Opt_serv_policy_opts *po = opt_policy_find(opts, defp, 1, defp->len);
    
    if (!po)
      vlg_warn(vlg, "policy -- %s -- can't be found\n",
               vstr_export_cstr_ptr(defp, 1, defp->len));
    else
      audispd_policy_change_con(con, (Audispd_policy_opts *)po);
  }
  
  vlg_info(vlg, "MANAGED-CONNECT from[$<sa:%p>]\n", EVNT_SA(con->evnt));

  con->evnt->flag_fully_acpt = TRUE;
  
  return (TRUE);

 mk_tm_fail:
 init_con_end_fail:
 mk_limit_fail:
  evnt_close(con->evnt);
  return (FALSE);
  
 mk_evnt_fail:
 mk_sa_fail:
  vstr_free_base(con->tag);
  if ((con__beg = con->next))
    con__beg->prev = NULL;
  --audispd__io_num;
 init_con_fail:
  F(con);
 mk_con_fail:
  
 fork_fail:
  close(fds[0]);
  close(fds[1]);
 sockpair_fail:
  return (FALSE);
}

static void serv_cb_func_free(struct Evnt *evnt)
{
  struct Con *con = (struct Con *)evnt;

  opt_serv_sc_free_beg(evnt, "FREE");

  /*  ASSERT(con->fs && !con->fs_num && !con->fs_off && (con->fs->fd == -1)); */

  if (con->managed)
  {
    switch (con->policy->managed_con_died)
    {
      case AUDISPD_TYPE_MAN_CON_DIED_RESTART:
        if (audispd__input && serv_make_managed_con(con->managed))
          break;
        
        /* if it failed ... fall through and cascade... */
        /* FIXME: ... make this configurable? */
        
      case AUDISPD_TYPE_MAN_CON_DIED_CASCADE:
        evnt_close(audispd__input);
        break;
      case AUDISPD_TYPE_MAN_CON_DIED_IGNORE:
        
        ASSERT_NO_SWITCH_DEF();
    }
  }
  
  vstr_free_base(con->tag);

  --audispd__io_num;
  if (con->next)
    con->next->prev = con->prev;
  if (con->prev)
    con->prev->next = con->next;
  else
    con__beg        = con->next;
  
  F(con);
}

static const struct Evnt_cbs audispd__con_evnt_cbs =
{
 evnt_cb_func_accept,    /* def */
 evnt_cb_func_connect,   /* def */

 serv_cb_func_recv,
 serv_cb_func_send,

 serv_cb_func_free,
 evnt_cb_func_shutdown_r /* def */
};

static struct Evnt *serv_cb_func_accept(struct Evnt *from_evnt, int fd,
                                        struct sockaddr *sa, socklen_t len)
{
  struct Acpt_listener *acpt_listener = (struct Acpt_listener *)from_evnt;
  struct Con *con = MK(sizeof(struct Con));
  Opt_serv_opts *opts = audispd_opts->s;
  
  if (sa->sa_family != AF_LOCAL) /* only support Local atm. */
    goto sa_fail;

  if (!audispd_con_init_beg(con))
    goto con_tag_fail;
  
  (void)len;
  if (!evnt_make_acpt_ref(con->evnt, fd, from_evnt->sa_ref))
    goto make_acpt_fail;
  con->evnt->cbs = &audispd__con_evnt_cbs;
  
  if (!evnt_limit_dup(con->evnt, &opts->io_nslimit))
    goto evnt_fail; /* redone on policy changes... */
  
  if (!evnt_limit_add(con->evnt,  opts->ref_io_limit))
    goto evnt_fail; /* redone on policy changes... */
  
  if (!evnt_limit_dup(con->evnt, &opts->io_nslimit))
    goto evnt_fail;
  
  if (!evnt_limit_add(con->evnt,  opts->ref_io_limit))
    goto evnt_fail;

  if (!audispd_con_init_end(con, acpt_listener))
     goto evnt_fail;

  if (!evnt_sc_timeout_via_mtime(con->evnt,
                                 con->policy->s->idle_timeout * 1000))
    VLG_WARNNOMEM_GOTO(evnt_fail, (vlg, "timeout: %m\n"));

  if (!opt_serv_sc_acpt_end(con->policy->s, from_evnt, con->evnt))
    goto evnt_fail;

  ASSERT(!con->evnt->flag_q_closed);

  con->evnt->flag_io_wonly_tm = TRUE;
  
  //  vlg_dbg2(vlg, "con($<sa:%p>): congestion=$<sockopt.s:%d%d%d>\n",
  //           CON_CEVNT_SA(con), evnt_fd(con->evnt), IPPROTO_TCP, TCP_CONGESTION);
  
  return (con->evnt);
  
 evnt_fail:
  evnt_close(con->evnt);
  return (con->evnt);
  
 make_acpt_fail:
  vstr_free_base(con->tag);
 con_tag_fail:
  F(con);
  VLG_WARNNOMEM_RET(NULL, (vlg, "%s: %m\n", "accept"));
  
 sa_fail:
  F(con);
  VLG_WARNNOMEM_RET(NULL, (vlg, "%s: AuDISPD sa($<sa:%p>) == Local fail\n",
                           "accept", sa));
}

static void audispd_serv_cb_func_free(struct Evnt *evnt)
{ /* if the listening connection dies, kill the main recv'ing evnt */
  evnt_close(audispd__input);
  
  evnt_sc_serv_cb_func_acpt_free(evnt);
}

static const struct Evnt_cbs audispd__serv_evnt_cbs = 
{
 serv_cb_func_accept,
 evnt_cb_func_connect,   /* def */

 evnt_cb_func_recv,      /* def */
 evnt_cb_func_send,      /* def */

 audispd_serv_cb_func_free,
 evnt_cb_func_shutdown_r /* def */
};

static void serv_make_bind(const char *program_name)
{
  Opt_serv_opts *opts = audispd_opts->s;
  Opt_serv_addr_opts *addr = opts->addr_beg;
  
  while (addr)
  {
    const char *acpt_address = NULL;
    const char *acpt_filter_file = NULL;
    /* const char *acpt_cong = NULL; */
    struct Evnt *evnt = NULL;
    
    OPT_SC_EXPORT_CSTR(acpt_address,     addr->acpt_address,     FALSE,
                       "accept address");
    OPT_SC_EXPORT_CSTR(acpt_filter_file, addr->acpt_filter_file, FALSE,
                       "accept filter file");
    /*
    OPT_SC_EXPORT_CSTR(acpt_cong,        addr->acpt_cong,        FALSE,
                       "accept congestion mode");
    */
    
    if (VALLEQ(audispd_opts->main_input, acpt_address))
      errx(EXIT_FAILURE, "main-input and bind addresses are the same: %s",
           acpt_address);
    
    evnt = evnt_sc_serv_make_bind_local(acpt_address,
                                        addr->q_listen_len,
                                        addr->max_connections,
                                        acpt_filter_file);

    if (addr->def_policy->len)
    {
      Vstr_base *dp = addr->def_policy;
      Opt_serv_policy_opts *po = opt_policy_find(opts, dp, 1, dp->len);
      Acpt_listener *acpt_listener = (Acpt_listener *)evnt;

      if (!po)
        usage(program_name, EXIT_FAILURE,
              " The name of the accept policy isn't valid.\n");
        
      acpt_listener->def_policy = vstr_ref_add(po->ref);
    }
    
    evnt->cbs = &audispd__serv_evnt_cbs;

    addr = addr->next;
  }
}

struct Audisp_hdr 
{ /* copy of struct audit_dispatcher_header */
 uint32_t        vers;    /* The version of this protocol */
 uint32_t        hlen;   /* Header length */
 uint32_t        type;   /* Message type */
 uint32_t        size;   /* Size of data following the header */
};

static int audispd_main_cb_func_recv(struct Evnt *evnt)
{ /* main audit dispatcher socket ... input for everything */
  struct Con *scan = NULL;
  unsigned int num = 0;  
  Vstr_base *data = evnt->io_r;
  unsigned int ern = 0;
  struct Audisp_hdr hdr[1];
  
  ASSERT(!evnt->io_r_shutdown);
  ASSERT(evnt == audispd__input);

  if (!evnt_recv(evnt, &ern))
  {
    if (ern != VSTR_TYPE_SC_READ_FD_ERR_EOF)
    {
      vlg_dbg2(vlg, "RECV ERR from[$<sa:%p>]: %u\n", EVNT_SA(evnt), ern);
      goto con_cleanup;
    }
    if (!evnt_shutdown_r(evnt, TRUE))
      goto con_cleanup;

    return (FALSE);
  }

  while (data->len >= sizeof(struct Audisp_hdr))
  {
    vstr_export_buf(data, 1, sizeof(struct Audisp_hdr), hdr, sizeof(hdr));
    
    if ((hdr->vers != 0) || (hdr->hlen != sizeof(struct Audisp_hdr)))
    {
      vlg_warn(vlg,
               "RECV ERR from[$<sa:%p>] vers[%u] hsz[%u] msg-type[%u:$<au-msg_type2name:%u>] sz[%u]\n",
             EVNT_SA(evnt),
             (unsigned int)hdr->vers,
             (unsigned int)hdr->hlen,
             (unsigned int)hdr->type,
             (unsigned int)hdr->type,
             (unsigned int)hdr->size);

      return (FALSE);
    }
    
    if (hdr->size > (data->len - hdr->hlen))
      break; /* not enough data, yet */
    
    evnt_got_pkt(evnt);
    
    vlg_dbg2(vlg,
             "RECV from[$<sa:%p>] msg-type[%u:$<au-msg_type2name:%u>] sz[%u]: $<vstr:%p%zu%zu>\n",
             EVNT_SA(evnt),
             (unsigned int)hdr->type,
             (unsigned int)hdr->type,
             (unsigned int)hdr->size,
             data, (size_t)(hdr->hlen + 1), (size_t)hdr->size);

    ASSERT(!(num = 0));
    scan = con__beg;
    ASSERT(!scan || !scan->prev);
    for (; scan; scan = scan->next)
    {
      Vstr_base *out = scan->evnt->io_w;
      unsigned int fscan = 0;

      ASSERT(++num);
      ASSERT(!scan->next || (scan->next->prev == scan));
      
      if (scan->evnt->flag_q_closed || scan->flag_drop_msgs)
        continue;

      while (fscan < scan->policy->filt_num)
      {
        if (scan->policy->filts[fscan] == (int)hdr->type)
          break;
        ++fscan;
      }
      if (fscan && (fscan == scan->policy->filt_num))
        continue;
          
      evnt_put_pkt(scan->evnt);
      
      /* FIXME: move to copying ptrs only ... not all data */
      vstr_add_vstr(out, out->len, data, 1, hdr->hlen + hdr->size,
                    VSTR_TYPE_ADD_DEF);
      
      if (out->len > scan->policy->max_data)
        switch (scan->policy->limit_reached)
        {
          case AUDISPD_TYPE_LIM_REACHED_DROP:
            scan->flag_drop_msgs = TRUE;
            break;

          case AUDISPD_TYPE_LIM_REACHED_CLOSE:
            evnt_close(scan->evnt);

            ASSERT_NO_SWITCH_DEF();
        }
    }
    ASSERT(num == audispd__io_num);
    
    if (data->conf->malloc_bad)
      goto con_cleanup;

    /* FIXME: move to copying ptrs only ... not all data */
    audispd__io_done_num = 0;
    vstr_del(data, 1, hdr->hlen + hdr->size);
  }

  scan = con__beg;
  ASSERT(!scan || !scan->prev);
  for (; scan; scan = scan->next)
  {
    ASSERT(++num);
    ASSERT(!scan->next || (scan->next->prev == scan));
    
    if (scan->evnt->flag_q_closed || scan->flag_drop_msgs)
      continue;
    
    if (!evnt_send_add(scan->evnt, TRUE, AUDISPD_CONF_MAX_WAIT_SEND))
      /* FIXME: just kill this plugin ... respawn, if we can etc. */
      /* goto con_cleanup; */
      evnt_close(scan->evnt);
  } 

  if (data->conf->malloc_bad)
    goto con_cleanup;
  
  return (TRUE);
  
 con_cleanup:
  evnt->io_r->conf->malloc_bad = FALSE;
  evnt->io_w->conf->malloc_bad = FALSE;
    
  return (FALSE);
}

static void audispd_main_cb_func_free(struct Evnt *evnt)
{ /* main audit dispatcher socket ... input for everything, once this is gone
     give up */
  struct Con *scan = NULL;
  
  ASSERT(evnt == audispd__input);

  evnt_acpt_close_all();
  
  scan = con__beg;
  ASSERT(!scan || !scan->prev);
  while (scan)
  {
    ASSERT(!scan->next || (scan->next->prev == scan));

    if (!scan->evnt->io_w->len)
      evnt_close(scan->evnt);
    scan = scan->next;
  } 

  evnt_vlg_stats_info(evnt, "MAIN FREE");

  audispd__input = NULL;

  F(evnt);
}

static int audispd_main_cb_func_connect(struct Evnt *evnt)
{
  ASSERT(evnt == audispd__input);
  
  evnt_pkt_wait_r(evnt);
  
  return (TRUE);
}

static const struct Evnt_cbs audispd__main_evnt_cbs =
{
 evnt_cb_func_accept,    /* def */
 audispd_main_cb_func_connect,

 audispd_main_cb_func_recv,
 evnt_cb_func_send,      /* def */

 audispd_main_cb_func_free,
 evnt_cb_func_shutdown_r /* def */
};

static void serv_make_main_input(const char *program_name)
{
  int fd = -1;
  
  if (!(audispd__input = MK(sizeof(struct Evnt))))
    errno = ENOMEM, err(EXIT_FAILURE, "make stdin Evnt");

  if (!audispd_opts->main_input->len)
  { /* use stdin */
    struct sockaddr_un tmp_sun;
    struct sockaddr_un *saddr = NULL;
    Vstr_ref *sa = NULL;
    
    if ((fd = dup(STDIN_FILENO)) == -1)
      err(EXIT_FAILURE, "dup stdin");

    tmp_sun.sun_path[0] = 0;

    if (!(sa = vstr_ref_make_malloc(SUN_LEN(&tmp_sun) + 1 +
                                    strlen("managed-connection-<stdin>"))))
      err(EXIT_FAILURE, "make sa");
    saddr = sa->ptr;
    saddr->sun_family = AF_LOCAL;
    strcpy(saddr->sun_path, "managed-connection-<stdin>");

    if (!evnt_make_custom(audispd__input, fd, sa, POLLIN))
      errx(EXIT_FAILURE, "make custom stdin");
    vstr_ref_del(sa);
  }
  else
  {
    const char *main_input = NULL;

    OPT_SC_EXPORT_CSTR(main_input, audispd_opts->main_input, FALSE,
                       "main input socket");
    
    if (!evnt_make_con_local(audispd__input, main_input))
      err(EXIT_FAILURE, "make local %s", main_input);
    audispd__input->flag_q_pkt_move = FALSE;
    
    if (audispd__input->flag_q_none)
      audispd_main_cb_func_connect(audispd__input);
  }
  
  close(STDIN_FILENO);
  if (open("/dev/null", O_RDONLY) != STDIN_FILENO)
    err(EXIT_FAILURE, "open stdin");
  
  audispd__input->cbs = &audispd__main_evnt_cbs;
  audispd__input->flag_fully_acpt = TRUE;
}

static void serv_make_main_plugins(const char *program_name)
{
  unsigned int scan = 0;

  while (scan < audispd_opts->cmd_num)
  {
    Audispd_cmd *cmd = &audispd_opts->cmds[scan++];
    const char *tmp = NULL;
    
    OPT_SC_EXPORT_CSTR(tmp, cmd->name, TRUE, "managed-connection name");
    
    if (vstr_spn_cstr_chrs_fwd(cmd->name, 1, cmd->name->len,
                               "abcdefghijklmnopqrstuvwxyz"
                               "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                               "0123456789-_.#") != cmd->name->len)
    {
      warnx(" Managed connection name: %s", tmp);
      usage(program_name, EXIT_FAILURE,
            " The value for managed-connection name must only include number,"
            " alphabetic, dot '.', hash '#', underbar '_' and"
            " dash '-' characters.\n");
    }

    vlg_dbg1(vlg, "cmd name=%s\n", tmp);
    
    OPT_SC_EXPORT_CSTR(tmp, cmd->path, TRUE, "managed-connection path");

    if (cmd->arg_num > 0)
      OPT_SC_EXPORT_CSTR(tmp, cmd->args[0], FALSE, "managed-connection arg1");
    if (cmd->arg_num > 1)
      OPT_SC_EXPORT_CSTR(tmp, cmd->args[1], FALSE, "managed-connection arg2");
    if (cmd->arg_num > 2)
      OPT_SC_EXPORT_CSTR(tmp, cmd->args[2], FALSE, "managed-connection arg3");
    if (cmd->arg_num > 3)
      OPT_SC_EXPORT_CSTR(tmp, cmd->args[3], FALSE, "managed-connection arg4");
    if (cmd->arg_num > 4)
      OPT_SC_EXPORT_CSTR(tmp, cmd->args[3], FALSE, "managed-connection arg5");
    if (cmd->arg_num > 5)
      OPT_SC_EXPORT_CSTR(tmp, cmd->args[3], FALSE, "managed-connection arg6");
    if (cmd->arg_num > 6)
      OPT_SC_EXPORT_CSTR(tmp, cmd->args[3], FALSE, "managed-connection arg7");
    if (cmd->arg_num > 7)
      OPT_SC_EXPORT_CSTR(tmp, cmd->args[3], FALSE, "managed-connection arg8");

    if (!serv_make_managed_con(cmd))
      errx(EXIT_FAILURE, "serv_make_managed_con()");
  }
}

static int serv_def_conf(Vstr_base *out)
{ /* pretend we passed (if it works...):
   * -C /etc/audit/audispd.conf --conf-dir /etc/audit/audispd-conf.d ... */
  struct stat64 d_stat[1];
  static const char pc_file[] = PATH_SYSCONFDIR "/audispd/audispd.conf";
  static const char pc_plg_dir[] = PATH_SYSCONFDIR "/audispd/plugins.d";
  static const char pc_pol_dir[] = PATH_SYSCONFDIR "/audispd/policies.d";
  Opt_serv_opts *opts = audispd_opts->s;

  /* use AUDIT "conf format", dir is for plugin "conf files" */
  if ((stat64(pc_file, d_stat) != -1) &&
      !auditd_audispd_main_conf_parse_file(out, audispd_opts, pc_file))
    return (FALSE);
  if ((stat64(pc_plg_dir, d_stat) != -1) &&
      !opt_serv_sc_config_dir(out, opts, pc_plg_dir,
                              auditd_audispd_sc_cmd_conf_parse_dir_file))
    return (FALSE);  
  if ((stat64(pc_pol_dir, d_stat) != -1) &&
      !opt_serv_sc_config_dir(out, opts, pc_pol_dir,
                              auditd_audispd_sc_pol_conf_parse_dir_file))
    return (FALSE);
  
  return (TRUE);
}

#define POPT_NUM_NR_ARG(x, y) do {                                      \
      Opt_serv_policy_opts *popt_scan = audispd_opts->s->def_policy;    \
                                                                        \
      while (popt_scan)                                                 \
      {                                                                 \
        Audispd_policy_opts *popt_tmp = (Audispd_policy_opts *)popt_scan;   \
        OPT_NUM_NR_ARG(popt_tmp->x, y);                                 \
        popt_scan = popt_scan->next;                                    \
      }                                                                 \
    } while (FALSE)


static void serv_cmd_line(int argc, char *argv[])
{
  int optchar = 0;
  const char *program_name = NULL;
  struct option long_options[] =
  {
   OPT_SERV_DECL_GETOPTS(),
   
   {"local", required_argument, NULL, 'H'},
   
   {"configuration-file", required_argument, NULL, 'C'},
   {"config-file",        required_argument, NULL, 'C'},
   {"conf-file",          required_argument, NULL, 'C'},
   {"configuration-directory", required_argument, NULL, 140},
   {"config-dir",              required_argument, NULL, 140},
   {"conf-dir",                required_argument, NULL, 140},
   {"configuration-data-daemon", required_argument, NULL, 143},
   {"config-data-daemon",        required_argument, NULL, 143},
   {"configuration-data-audit-dispatcherd", required_argument, NULL, 144},
   {"configuration-data-audispd", required_argument, NULL, 144},
   {"config-data-audispd",        required_argument, NULL, 144},

   {"plugins",                    required_argument, NULL, 140},
   {"managed-connections",        required_argument, NULL, 140},
   {"policies",                   required_argument, NULL, 145},
   
   {"default-configuration", no_argument, NULL, 149},
   {"def-config",            no_argument, NULL, 149},
   
   {"max-header-sz", required_argument, NULL, 128},
   {"main-input", required_argument, NULL, 129},
   {"input", required_argument, NULL, 129},
   /* 130 */
   /* 131 */
   /* 132 */
   /* 133 */
   /* 134 */
   /* 135 */
   /* 136 */
   /* 137 */
   /* 138 */
   /* 139 */
   /* 140 -- config. dir above */
   /* 141 */
   /* 142 */
   /* 143 -- config data above */
   /* 144 -- config data above */
   /* 145 -- policies */
   /* 146 */
   /* 147 */
   /* 148 */
   /* 149 -- config data above */
   {NULL, 0, NULL, 0}
  };
  const char *chroot_dir = NULL;
  Vstr_base *out = vstr_make_base(NULL);
  Opt_serv_opts *opts = audispd_opts->s;
  Audispd_policy_opts *popts = NULL;
  
  if (!out)
    errno = ENOMEM, err(EXIT_FAILURE, "command line");

  /* evnt_opt_nagle = TRUE; */
  
  program_name = opt_program_name(argv[0], PACKAGE);
  opts->name_cstr = program_name;
  opts->name_len  = strlen(program_name);

  opts->make_policy = audispd_policy_make;
  opts->copy_policy = audispd_policy_copy;

  if (!audispd_conf_init(audispd_opts))
    errno = ENOMEM, err(EXIT_FAILURE, "options");
  
  if (!geteuid()) /* If root */
    vstr_add_cstr_ptr(opts->addr_beg->acpt_address, 0, AUDISPD_CONF_DEF_LISTEN);
  else /* uncommon unprived default */
    vstr_add_cstr_ptr(opts->addr_beg->acpt_address, 0,
                      AUDISPD_CONF_DEF_USR_LISTEN);

  opts->become_daemon = TRUE; /* change audispd def., due to way it runs */
  
  popts = (Audispd_policy_opts *)opts->def_policy;
  while ((optchar = getopt_long(argc, argv, "C:dhH:M:nP:t:V",
                                long_options, NULL)) != -1)
  {
    switch (optchar)
    {
      case '?': usage(program_name, EXIT_FAILURE, "");
      case 'h': usage(program_name, EXIT_SUCCESS, "");
        
      case 'V':
        vstr_add_fmt(out, 0, " %s version %s.\n", program_name, VERSION);
        
        if (io_put_all(out, STDOUT_FILENO) == IO_FAIL)
          err(EXIT_FAILURE, "write");
        
        exit (EXIT_SUCCESS);
      
      OPT_SERV_GETOPTS(opts);

      case 'C':
        if (!auditd_audispd_main_conf_parse_file(out, audispd_opts, optarg))
          goto out_err_conf_msg;
        break;
      case 140:
        if (!opt_serv_sc_config_dir(out, opts, optarg,
                                    auditd_audispd_sc_cmd_conf_parse_dir_file))
          goto out_err_conf_msg;
        break;
      case 143:
        if (!opt_serv_conf_parse_cstr(out, opts, optarg))
          goto out_err_conf_msg;
      break;
      case 144:
        if (!audispd_conf_parse_cstr(out, audispd_opts, optarg))
          goto out_err_conf_msg;
        break;
      case 145:
        if (!opt_serv_sc_config_dir(out, opts, optarg,
                                    auditd_audispd_sc_pol_conf_parse_dir_file))
          goto out_err_conf_msg;
        break;
      case 149:
        if (!serv_def_conf(out))
          goto out_err_conf_msg;
        break;
        
      case 128: POPT_NUM_NR_ARG(max_header_sz, "max header size"); break;
      case 129: OPT_VSTR_ARG(audispd_opts->main_input);
      
      ASSERT_NO_SWITCH_DEF();
    }
  }

  ASSERT(optind >= 1);
  if (optind <= 1) /* just program name, no options so use default... */
    if (!serv_def_conf(out))
      goto out_err_conf_msg;
  vstr_free_base(out); out = NULL;
  
  argc -= optind;
  argv += optind;

  if (argc > 0)
    usage(program_name, EXIT_FAILURE, " Too many arguments.\n");

  {
    const char *pid_file = NULL;
    
    /* FIXME: move somewhere else */
    if (!opts->poll_backend->len)
      evnt_poll = &EVNT_POLL_BE_DEF;
    else if (VALLEQ(opts->poll_backend, "poll"))
      evnt_poll = &evnt_poll_be_poll;
    else if (VALLEQ(opts->poll_backend, "epoll"))
      evnt_poll = &evnt_poll_be_epoll;
    else
      usage(program_name, EXIT_FAILURE,
            " Not specified a valid poll backend.\n");
    evnt_poll->init();
  
    vlg_syslog_facility_set(vlg, opts->syslog_facility);
    /* vlg_date_set(vlg, opts->vlg_date_fmt_type); */
    if (!opts->vlg_tweaked_size)
    {
      if (opts->become_daemon)      
        opts->vlg_size = OPT_SERV_CONF_VLG_SIZE_UNTWEAKED_DAEMON;
      else
        opts->vlg_size = OPT_SERV_CONF_VLG_SIZE_UNTWEAKED_CONSOLE;
    }
    vlg_size_set(vlg, opts->vlg_size);
    vlg_syslog_native_set(vlg, opts->vlg_syslog_native);
  
    OPT_SC_EXPORT_CSTR(pid_file,   opts->pid_file,   FALSE, "pid file");
    OPT_SC_EXPORT_CSTR(chroot_dir, opts->chroot_dir, FALSE, "chroot directory");
    
    if (opts->drop_privs)
    {
      OPT_SERV_SC_RESOLVE_UID(opts);
      OPT_SERV_SC_RESOLVE_GID(opts);
    }
  
    serv_make_main_input(program_name);
    
    serv_make_bind(program_name);
    
    serv_make_main_plugins(program_name);
    
    if (opts->become_daemon)
    {
      /* FIXME: should be better ... more config. options
         
        if (daemon(FALSE, FALSE) == -1)
        err(EXIT_FAILURE, "daemon");
      */
      
      chdir("/"); /* don't fork, or setsid */

      dup2(STDIN_FILENO, STDOUT_FILENO); /* STDIN was done in make_main_input */
      dup2(STDIN_FILENO, STDERR_FILENO);
      
      vlg_daemon(vlg, program_name);
    }
    
    /* NOTE: after daemon so don't use err() anymore ... */
    
    if (opts->rlim_as_call)
      opt_serv_sc_rlim_as_num(opts->rlim_as_num);
    if (opts->rlim_core_call)
      opt_serv_sc_rlim_core_num(opts->rlim_core_num);
    if (opts->rlim_file_call)
      opt_serv_sc_rlim_file_num(opts->rlim_file_num);
  
    if (pid_file)
      vlg_pid_file(vlg, pid_file);

    if (opts->cntl_file->len)
      cntl_make_file(vlg, opts->cntl_file);

    if (chroot_dir)
    { /* preload locale info. so syslog can log in localtime, this doesn't work
       * with glibc's interfaces ...
       * but we have our own and gmtime_r() / localime_r() etc. */
      time_t now = time(NULL);
      (void)localtime(&now);
      
      vlg_sc_bind_mount(chroot_dir);
    }
    
    if (chroot_dir && ((chroot(chroot_dir) == -1) || (chdir("/") == -1)))
      vlg_err(vlg, EXIT_FAILURE, "chroot(%s): %m\n", chroot_dir);
    
    if (opts->drop_privs)
    {
      if (opts->keep_cap_fowner && (PROC_CNTL_KEEPCAPS(TRUE) == -1))
      {
        vlg_warn(vlg, "prctl(%s, %s): %m\n", "PR_SET_KEEPCAPS", "TRUE");
        opts->keep_cap_fowner = FALSE;
      }
      
      opt_serv_sc_drop_privs(opts);
      
      if (opts->keep_cap_fowner)
      { /* use + instead of = because that's how getpcaps prints it,
         * however they do the same due to no input */
        cap_t caps = cap_from_text("cap_fowner+ep-i");

        if (!caps)
          vlg_err(vlg, EXIT_FAILURE, "cap_from_text(%s): %m\n",
                  "cap_fowner+pe-i");
        
        if (PROC_CNTL_KEEPCAPS(FALSE) == -1)
          vlg_err(vlg, EXIT_FAILURE,
                  "prctl(%s, %s): %m\n", "PR_SET_KEEPCAPS", "FALSE");
        
        if (cap_set_proc(caps) == -1)
          vlg_err(vlg, EXIT_FAILURE, "cap_set_proc(%s): %m\n",
                  "cap_fowner+ep-i");
        if (cap_free(caps) == -1)
          vlg_err(vlg, EXIT_FAILURE, "cap_free(): %m\n");
      }
    }

    if (FALSE && (opts->num_procs > 1))
    {
      int has_cntl = !!opts->cntl_file->len;
      cntl_sc_multiproc(vlg, opts->num_procs, has_cntl, opts->use_pdeathsig);
    }
    
    if (opts->make_dumpable && (PROC_CNTL_DUMPABLE(TRUE) == -1))
      vlg_warn(vlg, "prctl(%s, %s): %m\n", "PR_SET_DUMPABLE", "TRUE");
  }

  audispd_opts->beg_time = time(NULL);
  
  {
    struct Evnt *evnt = evnt_queue("accept");
    
    while (evnt)
    {
      vlg_info(vlg, "READY [$<sa:%p>]: %s\n", EVNT_SA(evnt),
               chroot_dir ? chroot_dir : "");
      evnt = evnt->next;
    }
  }

  opt_serv_conf_free_beg(opts);
  return;

 out_err_conf_msg:
  if (opts->become_daemon && !isatty(STDERR_FILENO))
  { /* we are mostly started from auditd, and it doesn't understand STDERR
     * so put error messages where people can see them */
    vlg_daemon(vlg, program_name);

    ASSERT(!vlg->out_vstr->len);
    
    vstr_mov(vlg->out_vstr, 0, out, 1, out->len);
    vlg_err(vlg, EXIT_FAILURE, "\n");
  }

  vstr_add_cstr_ptr(out, out->len, "\n");
  if (io_put_all(out, STDERR_FILENO) == IO_FAIL)
    err(EXIT_FAILURE, "write");
  exit (EXIT_FAILURE);  
}

#if COMPILE_DEBUG
static int cov(void) /* for coverage */
{
  Opt_serv_policy_opts *opt = NULL;
  
  assert((F(M0(1, 1)), 1));
  
  ASSERT(date_syslog_trad(vlg->dt, time(NULL)));

  opt = opt_policy_make(audispd_opts->s);
  vstr_ref_del(opt->ref);
  
  return (TRUE);
}
#endif

int main(int argc, char *argv[])
{
  if (sizeof(uintmax_t) != sizeof(VSTR_AUTOCONF_uintmax_t))
    errx(EXIT_FAILURE, "uintmax_t size is different between program and Vstr");

  serv_init();

  serv_cmd_line(argc, argv);

  ASSERT(cov()); /* for coverage */
  
  while (evnt_waiting())
  {
    evnt_sc_main_loop(AUDISPD_CONF_MAX_WAIT_SEND);

    { /* opt_serv_sc_check_children(); */

      /* FIXME: no MP atm. ... just managed-connections,
       *        handled via. their fds */
      if (evnt_signal_child_exited)
      {
        while (waitpid(-1, NULL, WNOHANG) > 0)
          /* do nothing ... */;
        evnt_signal_child_exited = FALSE;
      }

      /* HACK */
      if (evnt_signal_term)
      {
        vlg_info(vlg, "Termination signal received, closing down.\n");
        evnt_acpt_close_all();
        evnt_scan_q_close();
        evnt_signal_term = FALSE;
      }
    }
    
    opt_serv_sc_cntl_resources(audispd_opts->s);
  }
  evnt_out_dbg3("E");
  
  evnt_timeout_exit();
  cntl_child_free();
  evnt_close_all();

  /*  httpd_exit(); */
  
  vlg_free(vlg);
  vlg_exit();

  audispd_conf_free(audispd_opts);
  
  timer_q_exit();
  
  vstr_exit();

  MALLOC_CHECK_EMPTY();
  
  exit (EXIT_SUCCESS);
}
