
#define EX_UTILS_NO_FUNCS 1
#include "ex_utils.h"

#include "mk.h"

#include "plugin-audisp2file.h"

Opt_serv_policy_opts *audisp2file_policy_make(Opt_serv_opts *beg)
{
  Audisp2file_policy_opts *opts = MK(sizeof(Audisp2file_policy_opts));
  Vstr_ref *ref = NULL;
  
  if (!opts)
    goto mk_opts_fail;
  
  if (!(ref = vstr_ref_make_ptr(opts, opt_policy_free)))
    goto mk_ref_fail;
  opts->s->ref = ref;

  if (!opt_policy_init(beg, opts->s))
    goto policy_init_fail;

  /* audisp2file_policy_init() */
  
  return (opts->s);

 policy_init_fail:
  ref->ptr = NULL;
  vstr_ref_del(ref);
 mk_ref_fail:
  F(opts);
 mk_opts_fail:
  return (NULL);
}

void audisp2file_policy_change_con(struct Con *con,
                                   Audisp2file_policy_opts *policy)
{
  opt_policy_change_con(con->evnt, policy->s);
  con->policy = policy;
}

int audisp2file_policy_copy(Opt_serv_policy_opts *passed_dst,
                            const Opt_serv_policy_opts *passed_src)
{
  Audisp2file_policy_opts *dst = (Audisp2file_policy_opts *)passed_dst;
  const Audisp2file_policy_opts *src = (const Audisp2file_policy_opts *)passed_src;

  return (TRUE);
}

int  audisp2file_conf_init(Audisp2file_opts *audisp2file_opts)
{
  unsigned int scan = 0;
  
  if (!opt_serv_conf_init(audisp2file_opts->s, AF_LOCAL))
    goto opts_init_fail;
  
  if (!(audisp2file_opts->main_input = vstr_make_base(NULL)))
    goto audisp2file_init_fail;
  
  return (TRUE);
  
 audisp2file_init_fail:
  audisp2file_conf_free(audisp2file_opts);
 opts_init_fail:
  return (FALSE);
}

void audisp2file_conf_free(Audisp2file_opts *audisp2file_opts)
{
  Opt_serv_policy_opts *scan = audisp2file_opts->s->def_policy;
  unsigned int cscan = 0;
  
  while (FALSE && scan)
  {
    // Audisp2file_policy_opts *tmp = (Audisp2file_policy_opts *)scan;
    Opt_serv_policy_opts *scan_next = scan->next;

    /* nothing atm. */
    
    scan = scan_next;
  }

  opt_policy_sc_all_ref_del(audisp2file_opts->s);

  vstr_free_base(audisp2file_opts->main_input);
  
  conf_parse_free(audisp2file_opts->conf); audisp2file_opts->conf = NULL;

  opt_serv_conf_free_beg(audisp2file_opts->s);
  opt_serv_conf_free_end(audisp2file_opts->s);
}

static int audisp2file__conf_main_policy_d1(Audisp2file_policy_opts *opts,
                                        Conf_parse *conf, Conf_token *token,
                                        int clist)
{
  if (0) { }
  
  else if (OPT_SERV_SYM_EQ("match-init"))
    OPT_SERV_SC_MATCH_INIT(opts->s->beg,
                           audisp2file__conf_main_policy_d1(opts, conf, token,
                                                        clist));
    
  else
    return (FALSE);
  
  return (TRUE);
}

static int audisp2file__conf_main_policy(Audisp2file_opts *opts,
                                     Conf_parse *conf, Conf_token *token)
{
  Opt_serv_policy_opts *popts = NULL;
  Conf_token *ntoken = NULL;
  unsigned int cur_depth = opt_policy_sc_conf_parse(opts->s, conf, token,
                                                    &popts, &ntoken);
  
  if (!cur_depth)
    return (FALSE);

  do
  {
    int clist = FALSE;
    
    CONF_SC_MAKE_CLIST_MID(cur_depth, clist);
    
    else if (audisp2file__conf_main_policy_d1((Audisp2file_policy_opts *)popts,
                                          conf, token, clist))
    { }
    
    CONF_SC_MAKE_CLIST_END();
  } while (ntoken &&
           (cur_depth = opt_policy_sc_conf_parse(opts->s, conf, token,
                                                 &popts, &ntoken)));
  return (TRUE);
}

static int audisp2file__conf_main_d1(Audisp2file_opts *audisp2file_opts,
                                 Conf_parse *conf, Conf_token *token, int clist)
{
  if (OPT_SERV_SYM_EQ("org.and.daemon-conf-1.0"))
  {
    if (!opt_serv_conf(audisp2file_opts->s, conf, token))
      return (FALSE);
  }
  else if (OPT_SERV_SYM_EQ("match-init"))
    OPT_SERV_SC_MATCH_INIT(audisp2file_opts->s,
                           audisp2file__conf_main_d1(audisp2file_opts,
                                                 conf, token, clist));
  
  else if (OPT_SERV_SYM_EQ("main-input"))
  {
    return (opt_serv_sc_make_static_input(audisp2file_opts->s, conf, token,
                                          audisp2file_opts->main_input));
  }
  
  else if (OPT_SERV_SYM_EQ("policy"))
  {
    if (!audisp2file__conf_main_policy(audisp2file_opts, conf, token))
      return (FALSE);
  }
  
  /*  else if (OPT_SERV_SYM_EQ("match-connection"))
  {
    if (!audisp2file__match_make(conf, token, AUDISP2FILE_CONF_MAIN_MATCH_CON,
                             audisp2file_opts->match_connection,
                             audisp2file_opts->tmp_match_connection))
      return (FALSE);
    
    if (!conf_parse_end_token(conf, token, conf_token_at_depth(token)))
      return (FALSE);
  }
  */
  
  else
    return (FALSE);
  
  return (TRUE);  
}


static int audisp2file_conf_main(Audisp2file_opts *opts,
                             Conf_parse *conf, Conf_token *token)
{
  unsigned int cur_depth = token->depth_num;
  int clist = FALSE;
  
  if (!OPT_SERV_SYM_EQ("com.redhat.audisp2file-conf-main-1.0"))
    return (FALSE);

  CONF_SC_MAKE_CLIST_MID(cur_depth, clist);

  else if (audisp2file__conf_main_d1(opts, conf, token, clist))
  { }
  
  CONF_SC_MAKE_CLIST_END();
  
  /* And they all live together ... dum dum */
  if (conf->data->conf->malloc_bad)
    return (FALSE);
  
  return (TRUE);
}

#define AUDISP2FILE_CONF__BEG_APP() do {                                    \
      ASSERT(!opts->conf_num || (conf->state == CONF_PARSE_STATE_END)); \
      prev_conf_num = conf->sects->num;                                 \
      /* reinit */                                                      \
      conf->state = CONF_PARSE_STATE_BEG;                               \
    } while (FALSE)

/* restore the previous parsing we've done and skip parsing it again */
#define AUDISP2FILE_CONF__END_APP() do {                                    \
      if (opts->conf_num)                                               \
      {                                                                 \
        conf_parse_token(conf, token);                                  \
        conf_parse_num_token(conf, token, prev_conf_num);               \
      }                                                                 \
    } while (FALSE)

int audisp2file_conf_parse_cstr(Vstr_base *out, Audisp2file_opts *opts,const char *data)
{
  Conf_parse *conf    = opts->conf;
  Conf_token token[1] = {CONF_TOKEN_INIT};
  unsigned int prev_conf_num = 0;
  size_t pos = 1;
  size_t len = 0;

  ASSERT(opts && data);

  if (!conf && !(conf = conf_parse_make(NULL)))
    goto conf_malloc_fail;

  pos = conf->data->len + 1;
  if (!vstr_add_cstr_ptr(conf->data, conf->data->len,
                         "(com.redhat.audisp2file-conf-main-1.0 "))
    goto read_malloc_fail;
  if (!vstr_add_cstr_ptr(conf->data, conf->data->len, data))
    goto read_malloc_fail;
  if (!vstr_add_cstr_ptr(conf->data, conf->data->len,
                         ")"))
    goto read_malloc_fail;
  len = vstr_sc_posdiff(pos, conf->data->len);

  AUDISP2FILE_CONF__BEG_APP();
  
  if (!conf_parse_lex(conf, pos, len))
    goto conf_fail;

  AUDISP2FILE_CONF__END_APP();

  if (!conf_parse_token(conf, token))
    goto conf_fail;

  if ((token->type != CONF_TOKEN_TYPE_CLIST) || (token->depth_num != 1))
    goto conf_fail;
    
  if (!conf_parse_token(conf, token))
    goto conf_fail;

  ASSERT(OPT_SERV_SYM_EQ("com.redhat.audisp2file-conf-main-1.0"));
  if (!audisp2file_conf_main(opts, conf, token))
    goto conf_fail;

  if (token->num != conf->sects->num)
    goto conf_fail;

  opts->conf = conf;
  opts->conf_num++;
  
  return (TRUE);
  
 conf_fail:
  conf_parse_backtrace(out, data, conf, token);
 read_malloc_fail:
  conf_parse_free(conf);
 conf_malloc_fail:
  return (FALSE);
}

int audisp2file_conf_parse_file(Vstr_base *out, Audisp2file_opts *opts,
                            const char *fname)
{
  Conf_parse *conf    = opts->conf;
  Conf_token token[1] = {CONF_TOKEN_INIT};
  unsigned int prev_conf_num = 0;
  size_t pos = 1;
  size_t len = 0;
  
  ASSERT(opts && fname);

  if (!conf && !(conf = conf_parse_make(NULL)))
    goto conf_malloc_fail;

  pos = conf->data->len + 1;
  if (!vstr_sc_read_len_file(conf->data, conf->data->len, fname, 0, 0, NULL))
    goto read_malloc_fail;
  len = vstr_sc_posdiff(pos, conf->data->len);

  AUDISP2FILE_CONF__BEG_APP();
  
  if (!conf_parse_lex(conf, pos, len))
    goto conf_fail;

  AUDISP2FILE_CONF__END_APP();
  
  while (conf_parse_token(conf, token))
  {
    if ((token->type != CONF_TOKEN_TYPE_CLIST) || (token->depth_num != 1))
      goto conf_fail;
    
    if (!conf_parse_token(conf, token))
      goto conf_fail;
    
    if (OPT_SERV_SYM_EQ("org.and.daemon-conf-1.0"))
    {
      if (!opt_serv_conf(opts->s, conf, token))
        goto conf_fail;
    }
    else if (OPT_SERV_SYM_EQ("com.redhat.audisp2file-conf-main-1.0"))
    {
      if (!audisp2file_conf_main(opts, conf, token))
        goto conf_fail;
    }
    else
      goto conf_fail;
    
    opts->conf_num++;
  }

  opts->conf = conf;
  
  return (TRUE);
  
 conf_fail:
  conf_parse_backtrace(out, fname, conf, token);
  errno = 0;
 read_malloc_fail:
  if (errno && out) /* can't find config. file */
    vstr_add_fmt(out, out->len, "open(%s): %m", fname);
  
  conf_parse_free(conf);
 conf_malloc_fail:
  return (FALSE);
}

int audisp2file_sc_conf_parse_dir_file(Vstr_base *out, void *opts,
                                   const char *fname)
{
  return (audisp2file_conf_parse_file(out, opts, fname));
}
