#! /usr/bin/perl -w

use strict;
use File::Path;

my $clean_on_exit = 1;

$ENV{VSTR_TST_MP} = 0; # Only one input, so we are screwed...

push @INC, "$ENV{SRCDIR}/tst";
require 'vstr_tst_examples.pl';

our $tst_DBG;
our $root;

my $audispd__input = undef;

sub audisp_cntl_list
  { # FIXME: see if it looks "OK"
    my $list_pid = tst_proc_fork();
    if (!$list_pid) {
      sleep(2);
      system("and-cntl -e list audispd_cntl > /dev/null");
      system("and-cntl -e 'list foo' audispd_cntl > /dev/null");
      _exit(0);
    }
    return $list_pid;
  }

sub audispd_file_tst
  {
    my $io_r = shift;
    my $io_w = shift;
    my $xtra = shift || {};
    my $sz   = shift;

    my $data = daemon_get_io_r($io_r);

# open (D, "| xxd");
    $data =~ s/
	       Vers: \s (\d+)\n
               Type: \s (\d+)\n
               Data-beg: \s (.*?) \n
               Data-end: \s? \n+
	       /pack("I4", ($1, 16, $2, length($3))) . $3/egsx;

# print D $data;
    my $out_name = $io_r;
    if ($out_name !~ /_conf_(\d+)_(tst)_(\d+)/)
      { failure(" Bad test name."); }

    $out_name = "out_conf_audispd_${2}_${1}_${3}";

    open(my $audispd__output, "< $out_name") ||
      failure("Can't open output file.");
    my $output = daemon_io($data,
                           1, $xtra->{slow_write}, 1,
			   $audispd__input, $audispd__output);
    $audispd__output->close();

    daemon_put_io_w($io_w, $output);
  }

sub all_conf_x_tsts
  {
    my $num = shift;
    my $prevnum = $num - 1;
    daemon_status("audispd_cntl", "127.0.$prevnum.1");

#    open(my $dummy, "> ") ||
#      failure("Can't open output file.");
#    $dummy->close();

    my @outs = grep /out_conf_audispd_tst_\d+_\d+/, glob "out_*";

    for my $i (@outs)
      { system("cp", "/dev/null", $i); }

    sub_tst(\&audispd_file_tst, "audispd_conf_$num");
return;
    for my $i (@outs)
      { system("cp", "/dev/null", $i); }

    sub_tst(\&audispd_file_tst, "audispd_conf_$num",
            {slow_write => 1});
  }

my $conf_root = "audispd_conf_root";

sub cleanup
  {
    print "DBG($$): audispd_cleanup()\n" if ($tst_DBG > 0);

    rmtree([$conf_root]);

    my @outs = grep /out_conf_audispd_tst_\d+_\d+/, glob "out_*";

    for my $i (@outs)
      { unlink($i); }

    my @soks = grep /listen_conf_audispd_tst_\d+/, glob "listen_conf_*";

    for my $i (@soks)
      { unlink($i); }
  }

sub setup
  {
    cleanup();

    mkpath([$conf_root . "/conf.d"]);

    my @conf_audispd_tsts = glob("$ENV{_TSTDIR}/conf_audispd_tst_*");
    for (1..scalar(@conf_audispd_tsts))
      {
        my $src_conf = "../../$ENV{_TSTDIR}/conf_audispd_tst_$_";
        if ($_ <= 2)
          {
            symlink($src_conf, "$conf_root/conf.d/__99_$_.conf");
          }
        elsif ($_ <= 8)
          {
            symlink($src_conf, "$conf_root/conf.d/_$_.conf");
          }
        else
          {
            symlink($src_conf, "$conf_root/conf.d/$_.conf");
          }
      }
  }

if (1)
  {
# Do testing...
    setup();

    my $args = ' --daemon=false --configuration-data-daemon "rlimit CORE <unlimited>"';


    my @conf_audispd_tsts = glob("$ENV{_TSTDIR}/conf_audispd_tst_*");
    my $beg = $ENV{TST_BEG} || 1;
    my $end = $ENV{TST_END} || scalar(@conf_audispd_tsts);

    if (($beg == 1) && ($end == scalar(@conf_audispd_tsts)))
      {
        $args .= " --config-dir $conf_root/conf.d";
      }
    else
      {
        for ($beg..$end)
          { $args .= " -C $ENV{_TSTDIR}/conf_audispd_tst_$_"; }
      }

    pipe(STDIN, $audispd__input) || failure("Bad pipe.");

    daemon_init("audispd", $root, $args . " -d -d -d");

    my $list_pid = audisp_cntl_list();


    for my $tnum ($beg..$end)
      {
        if (0) {}
        elsif (grep $tnum, (1..3))
	  {
            all_conf_x_tsts($tnum);
	  }
        else
          { failure("Bad conf number."); }
      }

    daemon_exit();

    cleanup();
  }

success();

END {
  my $save_exit_code = $?;
  if ($clean_on_exit)
    { daemon_cleanup(); }
  $? = $save_exit_code;
}
