
die " No SRCDIR" if (!defined ($ENV{SRCDIR}));

$ENV{_TSTDIR}   = "$ENV{SRCDIR}/tst";
# $ENV{_TOOLSDIR} = "$ENV{SRCDIR}/../tools";
# $ENV{_MIMEDIR}  = $ENV{_TOOLSDIR};

use strict;

use File::Basename;
use File::Compare;

use POSIX;

use IO::Socket;
use IO::Handle;

our $tst_DBG      = $ENV{VSTR_TST_DBG};
    $tst_DBG      =  0 if (!defined ($tst_DBG));

my $tst_mp       = $ENV{VSTR_TST_MP};
   $tst_mp       =  1 if (!defined ($tst_mp));
my $tst_num_mp   = $ENV{VSTR_TST_NUM_MP};
   $tst_num_mp   =  2 if (!defined ($tst_num_mp));
   $tst_num_mp   =  1 if ($tst_num_mp < 1);
   $tst_num_mp   =  1 if ($tst_mp == 0);
my $tst_proc_lim = $ENV{VSTR_TST_PROC_LIM};
   $tst_proc_lim = 32 if (!defined ($tst_proc_lim));
   $tst_proc_lim =  2 if ($tst_proc_lim < 2);

my $tst_tmout    = $ENV{VSTR_TST_TMOUT};
   $tst_tmout    = (1 * $tst_proc_lim) if (!defined ($tst_tmout));
my $tst_tmtry    = $ENV{VSTR_TST_TMTRY};
   $tst_tmtry    = 2 if (!defined ($tst_tmtry));

my $child_proc = 0;

my $xit_success = 0;
my $xit_failure = 1;
my $xit_fail_ok = 77;

STDOUT->autoflush(1);
STDERR->autoflush(1);


sub failure
  {
    my $txt = shift;

    warn("FAILURE($$) $0: $txt\n");
    if ($child_proc)
      { _exit($xit_failure); }
    else
      { exit($xit_failure); }
  }

sub success
  {
    tst_proc_waitall();
    if ($child_proc)
      { _exit($xit_success); }
    else
      { exit($xit_success); }
  }

my @TST__PROCS = ();
sub tst_fork()
  {
    my $pid = fork();
    if (defined($pid) && !$pid)
      {	
	$child_proc   = 1;
	$tst_mp       = 0;
	$tst_num_mp   = 1;
	$tst_proc_lim = 1;
	@TST__PROCS = ();
      }

    return $pid;
  }

sub tst_proc_wait()
  {
    my $pid = shift @TST__PROCS;

    print "DBG($$): wait() = $pid\n" if ($tst_DBG > 0);

    if (waitpid($pid, 0) <= 0)
      { failure("waitpid($pid): $!"); }
    my $code = $?;
    # 13 seems to be some weird perl error code.
    if (($code != $xit_success) && ($code != 13))
      { failure("waitpid($pid) == $code"); }
  }
sub tst_proc_waittry()
  {
    my @TMP__PROCS = @TST__PROCS; @TST__PROCS = ();
    for my $pid (@TMP__PROCS)
      {
	if (waitpid($pid, WNOHANG) <= 0)
	  { push @TST__PROCS, $pid; next; }

	print "DBG($$): wait() = $pid\n" if ($tst_DBG > 0);

	my $code = $?;
	# 13 seems to be some weird perl error code.
	if (($code != $xit_success) && ($code != 13))
	  { failure("waitpid($pid) == $code"); }
      }
  }
sub tst_proc_waitall()
  {
    while (@TST__PROCS > 0)
      { tst_proc_wait(); }
  }
sub tst_proc_fork()
  {
    if (@TST__PROCS >= $tst_proc_lim)
      {	tst_proc_waittry(); }
    while (@TST__PROCS >= $tst_proc_lim)
      {
	tst_proc_wait();
      }

    print "DBG($$): BEG fork() = [" . @TST__PROCS . "]\n" if ($tst_DBG > 2);
    my $pid = tst_fork();

    if (!defined ($pid))
      { failure("fork: $!"); }

    if ($pid)
      {
	print "DBG($$): fork() = $pid\n" if ($tst_DBG > 0);
	push @TST__PROCS, $pid;
	print "DBG($$): END fork($pid) = [" . @TST__PROCS . "]\n"
	  if ($tst_DBG > 2);
      }
    return $pid;
  }

sub sub_tst
  {
    my $func   = shift;
    my $prefix = shift;
    my $xtra   = shift;
    my $dir    = $ENV{_TSTDIR};
    my @files  = <$dir/${prefix}_tst_*>;
    my $sz     = scalar @files;

    print "TST: ${prefix}\n";

    print "DBG($$): files=@files\n" if ($tst_DBG);
    @files = undef;

    if (!$sz)
      { failure("NO files: ${prefix} [$dir/${prefix}_tst_*]"); }

    for my $num (1..($sz * $tst_num_mp))
      {
	--$num; $num %= $sz; ++$num;
	my $loc_tst_mp = $tst_mp;
	if ($loc_tst_mp)
	  {
	    my $pid = tst_proc_fork();
	    next    if ($pid);
	  }

	if (! -f "$dir/${prefix}_tst_$num")
	  { failure("NO file: $dir/${prefix}_tst_$num"); }
	if (! -f "$dir/${prefix}_out_$num")
	  { failure("NO file $dir/${prefix}_out_$num"); }

	my $fsz = -s "$dir/${prefix}_out_$num";

	unlink("${prefix}_tmp_${num}_$$");
	print "DBG($$): $dir/${prefix}_tst_$num ${prefix}_tmp_${num}\n"
	  if ($tst_DBG);
	$func->("$dir/${prefix}_tst_$num", "${prefix}_tmp_${num}_$$",
		$xtra, $fsz);

	if (compare("$dir/${prefix}_out_$num", "${prefix}_tmp_${num}_$$") != 0)
	  {
	    rename("${prefix}_tmp_${num}_$$", "${prefix}_tmp_${num}");
	    failure("tst ${prefix} $num");
	  }
	unlink("${prefix}_tmp_${num}_$$");
	unlink("${prefix}_tmp_${num}");

	if ($loc_tst_mp)
	  { success(); }
      }
    tst_proc_waittry();
  }

sub run_tst
  {
    my $cmd    = shift;
    my $prefix = shift || $cmd;
    my $opts   = shift || "";

    sub sub_run_tst
      {
	my $io_r = shift;
	my $io_w = shift;
	my $xtra = shift;

	my $cmd  = $xtra->{cmd};
	my $opts = $xtra->{opts};

	system("./${cmd} $opts -- $io_r > $io_w");
      }

    sub_tst(\&sub_run_tst,      $prefix, {cmd => $cmd, opts => $opts});
  }

sub run_err_tst
  {
    my $cmd    = shift;
    my $prefix = shift || $cmd;
    my $opts   = shift || "";

    sub sub_run_err_tst
      {
	my $io_r = shift;
	my $io_w = shift;
	my $xtra = shift;

	my $cmd  = $xtra->{cmd};
	my $opts = $xtra->{opts};

	open(TMP, ">&STDERR")    || failure("Can't dup stderr.");
	open(STDERR, ">> $io_w") || failure("Can't dup stderr.");
	system("./${cmd} $opts -- $io_r > $io_w");
	open(STDERR, ">&TMP")    || failure("Can't dup stderr.");
      }

    sub_tst(\&sub_run_err_tst, $prefix, {cmd => $cmd, opts => $opts});
  }

sub run_simple_tst
  {
    my $cmd    = shift;
    my $prefix = shift || $cmd;

    sub sub_run_simple_tst
      {
	my $io_r = shift;
	my $io_w = shift;
	my $xtra = shift;

	my $cmd  = $xtra->{cmd};

	system("./${cmd} $io_r > $io_w");
      }

    sub_tst(\&sub_run_simple_tst, $prefix, {cmd => $cmd});
  }

sub run_dir_tst
  {
    my $cmd    = shift;
    my $dir    = shift;
    my $prefix = shift || $cmd;

    sub sub_run_dir_tst
      {
	my $io_r = shift;
	my $io_w = shift;
	my $xtra = shift;

	my $cmd  = $xtra->{cmd};
	my $dir  = $xtra->{dir};

	system("./${cmd} $dir > $io_w");
      }

    sub_tst(\&sub_run_dir_tst, $prefix, {cmd => $cmd, dir => $dir});
  }

{
my $pdaemon_pid = undef;
my $ldaemon_pid = undef;
my $daemon_pid  = undef;
my $daemon_addr = undef;
my $daemon_port = undef;
my $daemon_cntl = undef;
sub daemon_status
  {
    $daemon_cntl = shift;
    my $daemon_laddr = shift;

    open(INFO, "and-cntl -e status ${daemon_cntl} |") ||
      failure("Can't open control ${daemon_cntl}.");

    while (<INFO>)
      {
	/^STATUS: / || next;

	if (/from\[(\d+[.]\d+[.]\d+[.]\d+)@(\d+)\]/)
	  {
	    ($daemon_addr, $daemon_port) = ($1, $2);
	  }
	elsif (/from\[local@([^\]]+)\]/)
	  {
	    ($daemon_addr) = ($1, $2);
	  }
	else { next }

	/pid\[(\d+)\]$/ || next;
	$daemon_pid = $1;

	if (defined ($daemon_laddr) && ($daemon_addr ne $daemon_laddr))
	  { next; }

	if ($daemon_addr eq '0.0.0.0')
	  {
	    $daemon_addr = '127.0.0.1';
	  }
	last;
      }

    close(INFO) || failure("Problem with cntl ${daemon_cntl}.");
  }

sub daemon_init
  {
    my $cmd    = shift;

    my $args   = shift || '';
    my $opts   = shift || "";

    my $cntl = "--cntl-file=${cmd}_cntl";
    my $port = ''; # "--port=0"; # Rand

    my $dbg = "";
    my $no_out = ">/dev/null  2> /dev/null";
    if ($tst_DBG > 1)
      {
	$no_out = '';
	$dbg    = '-d';
	if ($tst_DBG > 2)
	  { $dbg    = '-d -d'; }
	if ($tst_DBG > 3)
	  { $dbg    = '-d -d -d'; }
      }

    unlink("${cmd}_cntl"); # So we don't try connecting to the old one
    if ($args ne '')
      { $args = "-- $args"; }
    print "TST: ${cmd} $opts $args\n";

    $ldaemon_pid = tst_fork();
    if (!defined ($ldaemon_pid))
      { failure("fork($daemon_pid): $!"); }

    if (!$ldaemon_pid)
      { # Child
	my $args = "./${cmd} $port $opts $cntl $dbg $args $no_out";
	if (system($args))
	  { failure("daemon($args): $?"); }
	success("daemon($cmd)");
      }
    $pdaemon_pid = $$;

    # Wait for it...
    my $num = 0;
    while (! -e "${cmd}_cntl")
      {
	if (++$num > (60 * 5))
	  { failure("Can't find ${cmd}_cntl file"); }
	sleep(1);
      }

    daemon_status("${cmd}_cntl");
  }

sub daemon_pid
  {
    return $daemon_pid;
  }
sub daemon_addr
  {
    return $daemon_addr;
  }
sub daemon_port
  {
    return $daemon_port;
  }

sub nonblock {
    my $socket = shift;
    my $val = shift;
    my $flags;

    if (!defined($val))
      { $val = 1; }

    $flags = fcntl($socket, F_GETFL, 0)
      or failure("Can't get flags for socket: $!");

    return if (!!($flags & O_NONBLOCK) == !!$val);

    if ($val)
      { $flags |= O_NONBLOCK; }
    else
      { $flags &= ~O_NONBLOCK; }

    fcntl($socket, F_SETFL, $flags)
      or failure("Can't make socket nonblocking: $!");
}
sub daemon_connect_tcp
  {
    my $try = shift || 1;

    my $beg = time;
    my $sock = IO::Socket::INET->new(PeerAddr => daemon_addr(),
				     PeerPort => daemon_port(),
				     Proto    => "tcp",
				     Type     => SOCK_STREAM,
				     Timeout  => $tst_tmout);
    my $end = time;
    $end -= $beg;
    printf("DBG($$): connect (%s:%s) took %us\n",
	   daemon_addr(), daemon_port(), $end) if (!$sock || ($tst_DBG > 1));

    if (!$sock && ($try < $tst_tmtry))
      { return daemon_connect_tcp($try + 1); }
    if (!$sock)
      { failure("connect: $@"); }

    nonblock($sock);
    print "DBG($$): created TCP sock " . $sock->fileno . "\n" if ($tst_DBG > 2);

    if ($child_proc)
      { # connection is the only thing with a timeout, so help out the scheduler
	POSIX::nice(40);
      }
    return $sock;
  }

sub daemon_connect_local
  {
    my $try  = shift || 1;

    my $beg = time;
    my $sock = IO::Socket::UNIX->new(Peer     => daemon_addr(),
				     Type     => SOCK_STREAM,
				     Timeout  => $tst_tmout);
    my $end = time;
    $end -= $beg;
    printf("DBG($$): connect (%s:%s) took %us\n",
	   daemon_addr(), daemon_port(), $end) if (!$sock || ($tst_DBG > 1));

    if (!$sock && ($try < $tst_tmtry))
      { return daemon_connect_local($try + 1); }
    if (!$sock)
      { failure("connect: $@"); }

    nonblock($sock);
    print "DBG($$): created LOCAL sock " . $sock->fileno . "\n" 
      if ($tst_DBG > 2);

    if ($child_proc)
      { # connection is the only thing with a timeout, so help out the scheduler
	POSIX::nice(40);
      }
    return $sock;
  }

sub daemon_get_io_r
  {
    my $io_r = shift;

    my $data_r = "";
    { local ($/);
      open(IO_R, "< $io_r") || failure("open $io_r: $!");
      $data_r = <IO_R>;
      close(IO_R);
    }

    return $data_r;
  }

sub daemon_put_io_w
  {
    my $io_w   = shift;
    my $output = shift;

    open(IO_W, "> $io_w") || failure("open $io_w: $!");
    print IO_W $output;
    close(IO_W) || failure("close $io_w: $!");
  }

sub daemon_io
  {
    my $data = shift;
    my $done_shutdown = shift;
    my $slow_write = shift;
    my $allow_write_truncate = shift;
    my $len = length($data);
    my $sock_w = shift;
    my $sock_r = shift;
    my $off = 0;
    my $output = '';
    my $created_sock = 0;

    if (!defined ($sock_w))
      {
	$created_sock = 1;
	$sock_r = $sock_w = daemon_connect_tcp();
      }

    # don't do shutdown by default....
    if (!defined($done_shutdown)) { $done_shutdown = 1; }

    my $fin_write = !$len;
    while (1)
      {
	if ($len)
	  {
	    { # Block, kind of...
	      my $rin = '';
	      my $win = '';
	      vec($rin,fileno($sock_r),1) = 1;
	      vec($win,fileno($sock_w),1) = 1;
	      my $ein = $rin | $win;
	      select($rin, $win, $ein, 0.25);
	    }

	    my $wret = $len;

	    if ($slow_write && ($wret > 1))
	      { $wret /= 2; }
	    $wret = $sock_w->syswrite($data, $wret, $off);
	    if (!defined($wret) && $allow_write_truncate &&
		($! eq 'Connection reset by peer'))
	      { $wret = $len; }
	    if (!defined($wret) && ($! ne 'Resource temporarily unavailable'))
	      { failure("write: $!"); }
	    if (defined($wret))
	      {
		$len -= $wret; $off += $wret;
		if (!$len)
		  { $fin_write = 1; }
	      }
	  }
	if ($fin_write)
	  {
	    nonblock($sock_w, 0);
	    if (!$done_shutdown)
	      {
		$sock_w->shutdown(1)
		  if ($created_sock);
		$done_shutdown = 1;
	      }
	    $fin_write = 0;
	  }

	my $got_EOF = 0;
      try_again:
	my $ret = $sock_r->sysread(my($buff), 4096);

	next if (!defined($ret) &&
		 ($! eq 'Resource temporarily unavailable'));
	last if (!defined($ret) &&
		 ($! eq 'Connection reset by peer'));
	last if (!defined($ret) &&
		 ($! eq 'Connection reset by peer'));
	failure ("read: $!") if (!defined($ret));
	if (!$ret)
	  {
	    last if $got_EOF;

	    $got_EOF = 1;
	    sleep(2);
	    goto try_again;
	  }

	$output .= $buff;
      }

    if ($created_sock)
      {
	print "DBG($$): closed sock " . $sock_w->fileno . "\n" 
	  if ($tst_DBG > 2);
	$sock_w->close();
      }

    return $output;
}

sub daemon_exit
  {
    tst_proc_waitall();
    return if ($$ != $pdaemon_pid);

    if (system("and-cntl -e close $daemon_cntl > /dev/null"))
      { warn "cntl close: $!\n"; }
    unlink($daemon_cntl);
    $daemon_addr = undef;
    $daemon_port = undef;
    $daemon_cntl = undef;
    if (defined ($ldaemon_pid))
      {
	if (waitpid($ldaemon_pid, 0) <= 0)
	  { warn "waitpid(ldaemon[$ldaemon_pid]): $!\n"; }

	my $code = $?;
	if ($code != $xit_success)
	  { warn "waitpid(ldaemon[$ldaemon_pid]) == $code\n"; }
      }
  }
sub daemon_cleanup
  {
    if (defined($daemon_cntl) && -e $daemon_cntl)
      {
	daemon_exit()
      }
  }
}

1;
